"""Real signature module."""

from django.core.exceptions import ValidationError
from django.http import HttpRequest, HttpResponseBadRequest, HttpResponseNotFound, JsonResponse
from django.shortcuts import redirect

from appel_crises.data import POSTAL_CODE_TO_DISTRICT, DISTRICT_TO_MP
from appel_crises.mailer import send_mail, MAIN_PERSON, mailing_person
from appel_crises.models import Signature


def sign(request: HttpRequest):
    """Main signature function"""
    try:
        first_name = request.POST['firstname']
        surname = request.POST['surname']
        email = request.POST['email']
    except KeyError:
        return HttpResponseBadRequest()

    signature = Signature(
        first_name=first_name,
        surname=surname,
        email=email,
    )

    try:
        # Validate, among other things, the uniqueness of the email
        signature.full_clean()
    except ValidationError as e:
        # Probably a duplicate email, or a wrong email, or too long values
        return HttpResponseBadRequest(e)

    # I would put a try catch, but at this point i don't know what could go wrong
    signature.save()

    send_mail(
        sender=MAIN_PERSON,
        to=(mailing_person(signature.email, signature.first_name + ' ' + signature.surname),),
        html_content="<p>Thank you for signing</p>",
        text_content="Thank you for signing",
        subject="Thank you for signin"
    )

    return redirect('success')


def search_depute(request: HttpRequest):
    """Take input as POST parameter and return the corresponding depute as JSON"""

    postal_code = request.POST['postal_code']

    if postal_code not in POSTAL_CODE_TO_DISTRICT:
        return HttpResponseNotFound()

    districts = POSTAL_CODE_TO_DISTRICT[postal_code]

    mp_data = [DISTRICT_TO_MP[d] for d in districts if d in DISTRICT_TO_MP]

    # We have to pass safe=False, because we want to serialize an Array, see:
    # https://docs.djangoproject.com/en/3.0/ref/request-response/#serializing-non-dictionary-objects
    return JsonResponse(mp_data, safe=False)



"""Python module that manages the emails for the app.

It uses SendinBlue API as documented here : https://developers.sendinblue.com/reference

"""
from datetime import datetime
from typing import Dict, Iterable, Tuple, List

import requests

from appel_crises.settings import SENDIN_BLUE_API_KEY, MAIN_EMAIL, MAIN_NAME

SB_SEND_MAIL_URL = "https://api.sendinblue.com/v3/smtp/email"

SB_FETCH_EVENTS_URL = "https://api.sendinblue.com/v3/smtp/statistics/events"

SB_POST_HEADERS = {
    'accept': "application/json",
    'content-type': "application/json",
    'api-key': SENDIN_BLUE_API_KEY
}

SB_GET_HEADERS = {
    'accept': "application/json",
    'api-key': SENDIN_BLUE_API_KEY
}


MailingPerson = Dict[str, str]


def mailing_person(email: str, name: str = None) -> MailingPerson:
    """Build a MailingPerson dict."""
    if name is not None:
        return dict(name=name, email=email)
    else:
        return dict(email=email)


MAIN_PERSON = mailing_person(MAIN_EMAIL, MAIN_NAME)


def send_mail(sender: MailingPerson,
              to: Iterable[MailingPerson],
              html_content: str,
              subject: str,
              text_content: str = "",
              bcc: Iterable[MailingPerson] = None,
              cc: Iterable[MailingPerson] = None):
    """Uses SendinBlue API to send an email"""

    body = dict(
        sender=sender,
        to=to,
        htmlContent=html_content,
        textContent=text_content,
        subject=subject,
    )

    if bcc is not None:
        body['bcc'] = bcc

    if cc is not None:
        body['cc'] = cc

    response = requests.post(SB_SEND_MAIL_URL, headers=SB_POST_HEADERS, json=body)

    return response.json()


def get_delivered_emails(offset: int, limit: int = 100) -> List[Tuple[str, datetime]]:
    """Fetch the delivered emails."""
    params = dict(offset=offset, limit=limit, event="delivered")

    response = requests.get(SB_FETCH_EVENTS_URL, headers=SB_GET_HEADERS, params=params)

    data = response.json()
    if 'events' not in data:
        return []

    result = [(e['email'], datetime.fromisoformat(e['date'])) for e in data['events']]

    return result



"""Module that provides the views of the app."""
from django.http import HttpRequest
from django.shortcuts import render

from appel_crises.models import Signature


def home(request: HttpRequest):
    """Home view."""
    context = dict(
        counter=Signature.get_count(),
    )

    return render(request, "home.html", context=context)


def success(request: HttpRequest):
    """Signature success page"""
    return render(request, "success.html")


def form_no_js(request: HttpRequest):
    return render(request, "form_no_js.html")



from datetime import datetime

from django.db import models
from django.db.models import Max
from django.utils import timezone


class Signature(models.Model):
    created_at = models.DateTimeField(default=timezone.now)
    email = models.EmailField(max_length=100, unique=True)
    first_name = models.CharField(max_length=100)
    surname = models.CharField(max_length=100)
    email_sent_at = models.DateTimeField(null=True, blank=True, db_index=True)
    verified_at = models.DateTimeField(null=True, blank=True, db_index=True)

    def __str__(self):
        return f"{self.first_name} {self.surname} {'(verified) ' if self.verified_at else ''}: " \
               f"{self.email} "

    @staticmethod
    def get_count() -> int:
        """Returns the number of verified signatures."""
        return Signature.objects.filter(verified_at__isnull=False).count()

    @staticmethod
    def mark_email_as_verified(email: str, date: datetime):
        """Mark email as verified."""
        try:
            signature = Signature.objects.get(email=email)
        except Signature.DoesNotExist:
            return False

        signature.verified_at = date
        signature.save()

        return True


class EmailOffset(models.Model):
    """Stores the last offset of events fetched by get_delivered_emails."""
    created_at = models.DateTimeField(default=timezone.now)
    offset = models.IntegerField(default=0, db_index=True)

    @staticmethod
    def get_max_offset() -> int:
        """Return the maximum offset stored in the table."""
        result = EmailOffset.objects.aggregate(Max('offset'))['offset__max']

        if result is None:
            return 0
        else:
            return result
